from unittest import mock
from zope.interface import alsoProvides
from twisted.internet.task import Clock
from twisted.internet.interfaces import ITransport
import pytest_twisted

from ...eventual import EventualQueue
from ..._interfaces import IDilationConnector
from ..._dilation.roles import LEADER, FOLLOWER
from ..._dilation.connection import (DilatedConnectionProtocol, encode_record,
                                     KCM, Open, Ack)
from .common import clear_mock_calls


def make_con(role, use_relay=False):
    clock = Clock()
    eq = EventualQueue(clock)
    connector = mock.Mock()
    alsoProvides(connector, IDilationConnector)
    n = mock.Mock()  # pretends to be a Noise object
    n.write_message = mock.Mock(side_effect=[b"handshake"])
    c = DilatedConnectionProtocol(eq, role, "desc", connector, n,
                                  b"outbound_prologue\n", b"inbound_prologue\n")
    if use_relay:
        c.use_relay(b"relay_handshake\n")
    t = mock.Mock()
    alsoProvides(t, ITransport)
    return c, n, connector, t, eq


def test_hashable():
    c, n, connector, t, eq = make_con(LEADER)
    hash(c)

@pytest_twisted.ensureDeferred
async def test_bad_prologue():
    c, n, connector, t, eq = make_con(LEADER)
    c.makeConnection(t)
    d = c.when_disconnected()
    assert n.mock_calls == [mock.call.start_handshake()]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    clear_mock_calls(n, connector, t)

    c.dataReceived(b"prologue\n")
    assert n.mock_calls == []
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.loseConnection()]

    eq.flush_sync()
    assert not d.called
    c.connectionLost(b"why")
    eq.flush_sync()
    conn = await d
    assert conn is c

def _test_no_relay(role):
    c, n, connector, t, eq = make_con(role)
    t_kcm = KCM()
    t_open = Open(seqnum=1, scid=0x11223344, subprotocol="proto")
    t_ack = Ack(resp_seqnum=2)
    n.decrypt = mock.Mock(side_effect=[
        encode_record(t_kcm),
        encode_record(t_open),
    ])
    exp_kcm = b"\x00\x00\x00\x03kcm"
    n.encrypt = mock.Mock(side_effect=[b"kcm", b"ack1"])
    m = mock.Mock()  # Manager

    c.makeConnection(t)
    assert n.mock_calls == [mock.call.start_handshake()]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    clear_mock_calls(n, connector, t, m)

    c.dataReceived(b"inbound_prologue\n")

    exp_handshake = b"\x00\x00\x00\x09handshake"
    if role is LEADER:
        # the LEADER sends the Noise handshake message immediately upon
        # receipt of the prologue
        assert n.mock_calls == [mock.call.write_message()]
        assert t.mock_calls == [mock.call.write(exp_handshake)]
    else:
        # however the FOLLOWER waits until receiving the leader's
        # handshake before sending their own
        assert n.mock_calls == []
        assert t.mock_calls == []
    assert connector.mock_calls == []

    clear_mock_calls(n, connector, t, m)

    c.dataReceived(b"\x00\x00\x00\x0Ahandshake2")
    if role is LEADER:
        # we're the leader, so we don't send the KCM right away
        assert n.mock_calls == [
            mock.call.read_message(b"handshake2")]
        assert connector.mock_calls == []
        assert t.mock_calls == []
        assert c._manager is None
    else:
        # we're the follower, so we send our Noise handshake, then
        # encrypt and send the KCM immediately
        assert n.mock_calls == [
            mock.call.read_message(b"handshake2"),
            mock.call.write_message(),
            mock.call.encrypt(encode_record(t_kcm)),
        ]
        assert connector.mock_calls == []
        assert t.mock_calls == [
            mock.call.write(exp_handshake),
            mock.call.write(exp_kcm)]
        assert c._manager is None
    clear_mock_calls(n, connector, t, m)

    c.dataReceived(b"\x00\x00\x00\x03KCM")
    # leader: inbound KCM means we add the candidate
    # follower: inbound KCM means we've been selected.
    # in both cases we notify Connector.add_candidate(), and the Connector
    # decides if/when to call .select()

    assert n.mock_calls == [mock.call.decrypt(b"KCM")]
    assert connector.mock_calls == [mock.call.add_candidate(c)]
    assert t.mock_calls == []
    clear_mock_calls(n, connector, t, m)

    # now pretend this connection wins (either the Leader decides to use
    # this one among all the candidates, or we're the Follower and the
    # Connector is reacting to add_candidate() by recognizing we're the
    # only candidate there is)
    c.select(m)
    assert c._manager is m
    if role is LEADER:
        # TODO: currently Connector.select_and_stop_remaining() is
        # responsible for sending the KCM just before calling c.select()
        # iff we're the LEADER, therefore Connection.select won't send
        # anything. This should be moved to c.select().
        assert n.mock_calls == []
        assert connector.mock_calls == []
        assert t.mock_calls == []
        # there's a call to .have_peer(...) but we don't really care?
        assert len(m.mock_calls) == 1##[])

        c.send_record(KCM())
        assert n.mock_calls == [
            mock.call.encrypt(encode_record(t_kcm)),
        ]
        assert connector.mock_calls == []
        assert t.mock_calls == [mock.call.write(exp_kcm)]
        assert len(m.mock_calls) == 1  ##[]) .have_peer() call
    else:
        # follower: we already sent the KCM, do nothing
        assert n.mock_calls == []
        assert connector.mock_calls == []
        assert t.mock_calls == []
        assert len(m.mock_calls) == 1  # have_peer()
    clear_mock_calls(n, connector, t, m)

    c.dataReceived(b"\x00\x00\x00\x04msg1")
    assert n.mock_calls == [mock.call.decrypt(b"msg1")]
    assert connector.mock_calls == []
    assert t.mock_calls == []
    assert m.mock_calls == [mock.call.got_record(t_open)]
    clear_mock_calls(n, connector, t, m)

    c.send_record(t_ack)
    exp_ack = b"\x06\x00\x00\x00\x02"
    assert n.mock_calls == [mock.call.encrypt(exp_ack)]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"\x00\x00\x00\x04ack1")]
    assert m.mock_calls == []
    clear_mock_calls(n, connector, t, m)

    c.disconnect()
    assert n.mock_calls == []
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.loseConnection()]
    assert m.mock_calls == []
    clear_mock_calls(n, connector, t, m)

def test_no_relay_leader():
    return _test_no_relay(LEADER)

def test_no_relay_follower():
    return _test_no_relay(FOLLOWER)

def test_relay():
    c, n, connector, t, eq = make_con(LEADER, use_relay=True)

    c.makeConnection(t)
    assert n.mock_calls == [mock.call.start_handshake()]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"relay_handshake\n")]
    clear_mock_calls(n, connector, t)

    c.dataReceived(b"ok\n")
    assert n.mock_calls == []
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    clear_mock_calls(n, connector, t)

    c.dataReceived(b"inbound_prologue\n")
    assert n.mock_calls == [mock.call.write_message()]
    assert connector.mock_calls == []
    exp_handshake = b"\x00\x00\x00\x09handshake"
    assert t.mock_calls == [mock.call.write(exp_handshake)]
    clear_mock_calls(n, connector, t)


@pytest_twisted.ensureDeferred
async def test_relay_jilted():
    c, n, connector, t, eq = make_con(LEADER, use_relay=True)
    d = c.when_disconnected()

    c.makeConnection(t)
    assert n.mock_calls == [mock.call.start_handshake()]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"relay_handshake\n")]
    clear_mock_calls(n, connector, t)

    c.connectionLost(b"why")
    eq.flush_sync()
    conn = await d
    assert conn is c

def test_relay_bad_response():
    c, n, connector, t, eq = make_con(LEADER, use_relay=True)

    c.makeConnection(t)
    assert n.mock_calls == [mock.call.start_handshake()]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"relay_handshake\n")]
    clear_mock_calls(n, connector, t)

    c.dataReceived(b"not ok\n")
    assert n.mock_calls == []
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.loseConnection()]
    clear_mock_calls(n, connector, t)

def test_follower_combined():
    c, n, connector, t, eq = make_con(FOLLOWER)
    t_kcm = KCM()
    t_open = Open(seqnum=1, scid=0x11223344, subprotocol="proto")
    n.decrypt = mock.Mock(side_effect=[
        encode_record(t_kcm),
        encode_record(t_open),
    ])
    exp_kcm = b"\x00\x00\x00\x03kcm"
    n.encrypt = mock.Mock(side_effect=[b"kcm", b"ack1"])
    m = mock.Mock()  # Manager

    c.makeConnection(t)
    assert n.mock_calls == [mock.call.start_handshake()]
    assert connector.mock_calls == []
    assert t.mock_calls == [mock.call.write(b"outbound_prologue\n")]
    clear_mock_calls(n, connector, t, m)

    c.dataReceived(b"inbound_prologue\n")

    exp_handshake = b"\x00\x00\x00\x09handshake"
    # however the FOLLOWER waits until receiving the leader's
    # handshake before sending their own
    assert n.mock_calls == []
    assert t.mock_calls == []
    assert connector.mock_calls == []

    clear_mock_calls(n, connector, t, m)

    c.dataReceived(b"\x00\x00\x00\x0Ahandshake2")
    # we're the follower, so we send our Noise handshake, then
    # encrypt and send the KCM immediately
    assert n.mock_calls == [
        mock.call.read_message(b"handshake2"),
        mock.call.write_message(),
        mock.call.encrypt(encode_record(t_kcm)),
    ]
    assert connector.mock_calls == []
    assert t.mock_calls == [
        mock.call.write(exp_handshake),
        mock.call.write(exp_kcm)]
    assert c._manager is None
    clear_mock_calls(n, connector, t, m)

    # the leader will select a connection, send the KCM, and then
    # immediately send some more data

    kcm_and_msg1 = (b"\x00\x00\x00\x03KCM" +
                    b"\x00\x00\x00\x04msg1")
    c.dataReceived(kcm_and_msg1)

    # follower: inbound KCM means we've been selected.
    # in both cases we notify Connector.add_candidate(), and the Connector
    # decides if/when to call .select()

    assert n.mock_calls == [mock.call.decrypt(b"KCM"),
                                    mock.call.decrypt(b"msg1")]
    assert connector.mock_calls == [mock.call.add_candidate(c)]
    assert t.mock_calls == []
    clear_mock_calls(n, connector, t, m)

    # now pretend this connection wins (either the Leader decides to use
    # this one among all the candidates, or we're the Follower and the
    # Connector is reacting to add_candidate() by recognizing we're the
    # only candidate there is)
    c.select(m)
    assert c._manager is m
    # follower: we already sent the KCM, do nothing
    assert n.mock_calls == []
    assert connector.mock_calls == []
    assert t.mock_calls == []
    assert m.mock_calls[1:] == [mock.call.got_record(t_open)]
    clear_mock_calls(n, connector, t, m)



from itertools import count

from twisted.internet import reactor
from twisted.internet.threads import deferToThread

from unittest import mock

from pytest_twisted import ensureDeferred

from .._rlcompleter import (CodeInputter, _input_code_with_completion,
                            input_with_completion, warn_readline)
from ..errors import AlreadyInputNameplateError, KeyFormatError
import pytest

APPID = "appid"


@ensureDeferred
async def test_wrapper():
    helper = object()
    trueish = object()
    with mock.patch(
            "wormhole._rlcompleter._input_code_with_completion",
            return_value=trueish) as m:
        used_completion = await input_with_completion(
            "prompt:", helper, reactor)
    assert used_completion is trueish
    assert m.mock_calls == [mock.call("prompt:", helper, reactor)]
    # note: if this test fails, the warn_readline() message will probably
    # get written to stderr


# exercise _input_code_with_completion, which uses the blocking builtin
# "input()" function, hence _input_code_with_completion is usually in a
# thread with deferToThread

@mock.patch("wormhole._rlcompleter.CodeInputter")
@mock.patch("wormhole._rlcompleter.readline", __doc__="I am GNU readline")
@mock.patch("wormhole._rlcompleter.input", return_value="code")
def test_readline(input, readline, ci):
    c = mock.Mock(name="inhibit parenting")
    c.completer = object()
    trueish = object()
    c.used_completion = trueish
    ci.configure_mock(return_value=c)
    prompt = object()
    input_helper = object()
    reactor = object()
    used = _input_code_with_completion(prompt, input_helper, reactor)
    assert used is trueish
    assert ci.mock_calls == [mock.call(input_helper, reactor)]
    assert c.mock_calls == [mock.call.finish("code")]
    assert input.mock_calls == [mock.call(prompt)]
    assert readline.mock_calls == [
        mock.call.parse_and_bind("tab: complete"),
        mock.call.set_completer(c.completer),
        mock.call.set_completer_delims(""),
    ]

@mock.patch("wormhole._rlcompleter.CodeInputter")
@mock.patch("wormhole._rlcompleter.readline")
@mock.patch("wormhole._rlcompleter.input", return_value="code")
def test_readline_no_docstring(input, readline, ci):
    del readline.__doc__  # when in doubt, it assumes GNU readline
    c = mock.Mock(name="inhibit parenting")
    c.completer = object()
    trueish = object()
    c.used_completion = trueish
    ci.configure_mock(return_value=c)
    prompt = object()
    input_helper = object()
    reactor = object()
    used = _input_code_with_completion(prompt, input_helper, reactor)
    assert used is trueish
    assert ci.mock_calls == [mock.call(input_helper, reactor)]
    assert c.mock_calls == [mock.call.finish("code")]
    assert input.mock_calls == [mock.call(prompt)]
    assert readline.mock_calls == [
        mock.call.parse_and_bind("tab: complete"),
        mock.call.set_completer(c.completer),
        mock.call.set_completer_delims(""),
    ]

@mock.patch("wormhole._rlcompleter.CodeInputter")
@mock.patch("wormhole._rlcompleter.readline", __doc__="I am libedit")
@mock.patch("wormhole._rlcompleter.input", return_value="code")
def test_libedit(input, readline, ci):
    c = mock.Mock(name="inhibit parenting")
    c.completer = object()
    trueish = object()
    c.used_completion = trueish
    ci.configure_mock(return_value=c)
    prompt = object()
    input_helper = object()
    reactor = object()
    used = _input_code_with_completion(prompt, input_helper, reactor)
    assert used is trueish
    assert ci.mock_calls == [mock.call(input_helper, reactor)]
    assert c.mock_calls == [mock.call.finish("code")]
    assert input.mock_calls == [mock.call(prompt)]
    assert readline.mock_calls == [
        mock.call.parse_and_bind("bind ^I rl_complete"),
        mock.call.set_completer(c.completer),
        mock.call.set_completer_delims(""),
    ]

@mock.patch("wormhole._rlcompleter.CodeInputter")
@mock.patch("wormhole._rlcompleter.readline", None)
@mock.patch("wormhole._rlcompleter.input", return_value="code")
def test_no_readline(input, ci):
    c = mock.Mock(name="inhibit parenting")
    c.completer = object()
    trueish = object()
    c.used_completion = trueish
    ci.configure_mock(return_value=c)
    prompt = object()
    input_helper = object()
    reactor = object()
    used = _input_code_with_completion(prompt, input_helper, reactor)
    assert used is trueish
    assert ci.mock_calls == [mock.call(input_helper, reactor)]
    assert c.mock_calls == [mock.call.finish("code")]
    assert input.mock_calls == [mock.call(prompt)]

@mock.patch("wormhole._rlcompleter.CodeInputter")
@mock.patch("wormhole._rlcompleter.readline", None)
@mock.patch("wormhole._rlcompleter.input", return_value=b"code")
def test_bytes(input, ci):
    c = mock.Mock(name="inhibit parenting")
    c.completer = object()
    trueish = object()
    c.used_completion = trueish
    ci.configure_mock(return_value=c)
    prompt = object()
    input_helper = object()
    reactor = object()
    used = _input_code_with_completion(prompt, input_helper, reactor)
    assert used is trueish
    assert ci.mock_calls == [mock.call(input_helper, reactor)]
    assert c.mock_calls == [mock.call.finish("code")]
    assert input.mock_calls == [mock.call(prompt)]


def get_completions(c, prefix):
    completions = []
    for state in count(0):
        text = c.completer(prefix, state)
        if text is None:
            return completions
        completions.append(text)


def fake_blockingCallFromThread(f, *a, **kw):
    return f(*a, **kw)


def test_simple():
    # no actual completion
    helper = mock.Mock()
    c = CodeInputter(helper, "reactor")
    c.bcft = fake_blockingCallFromThread
    c.finish("1-code-ghost")
    assert not c.used_completion
    assert helper.mock_calls == [
        mock.call.choose_nameplate("1"),
        mock.call.choose_words("code-ghost")
    ]

@mock.patch(
    "wormhole._rlcompleter.readline",
    get_completion_type=mock.Mock(return_value=0))
def test_call(readline):
    # check that it calls _commit_and_build_completions correctly
    helper = mock.Mock()
    c = CodeInputter(helper, "reactor")
    c.bcft = fake_blockingCallFromThread

    # pretend nameplates: 1, 12, 34

    # first call will be with "1"
    cabc = mock.Mock(return_value=["1", "12"])
    c._commit_and_build_completions = cabc

    assert get_completions(c, "1") == ["1", "12"]
    assert cabc.mock_calls == [mock.call("1")]

    # then "12"
    cabc.reset_mock()
    cabc.configure_mock(return_value=["12"])
    assert get_completions(c, "12") == ["12"]
    assert cabc.mock_calls == [mock.call("12")]

    # now we have three "a" words: "and", "ark", "aaah!zombies!!"
    cabc.reset_mock()
    cabc.configure_mock(return_value=["aargh", "ark", "aaah!zombies!!"])
    assert get_completions(c, "12-a") == ["aargh", "ark", "aaah!zombies!!"]
    assert cabc.mock_calls == [mock.call("12-a")]

    cabc.reset_mock()
    cabc.configure_mock(return_value=["aargh", "aaah!zombies!!"])
    assert get_completions(c, "12-aa") == ["aargh", "aaah!zombies!!"]
    assert cabc.mock_calls == [mock.call("12-aa")]

    cabc.reset_mock()
    cabc.configure_mock(return_value=["aaah!zombies!!"])
    assert get_completions(c, "12-aaa") == ["aaah!zombies!!"]
    assert cabc.mock_calls == [mock.call("12-aaa")]

    c.finish("1-code")
    assert c.used_completion

def test_wrap_error():
    helper = mock.Mock()
    c = CodeInputter(helper, "reactor")
    c._wrapped_completer = mock.Mock(side_effect=ValueError("oops"))
    with mock.patch("wormhole._rlcompleter.traceback") as traceback:
        with mock.patch("wormhole._rlcompleter.print") as mock_print:
            with pytest.raises(ValueError) as e:
                c.completer("text", 0)
    assert traceback.mock_calls == [mock.call.print_exc()]
    assert mock_print.mock_calls == \
                     [mock.call("completer exception: oops")]
    assert str(e.value) == "oops"

@ensureDeferred
async def test_build_completions():
    rn = mock.Mock()
    # InputHelper.get_nameplate_completions returns just the suffixes
    gnc = mock.Mock()  # get_nameplate_completions
    cn = mock.Mock()  # choose_nameplate
    gwc = mock.Mock()  # get_word_completions
    cw = mock.Mock()  # choose_words
    helper = mock.Mock(
        refresh_nameplates=rn,
        get_nameplate_completions=gnc,
        choose_nameplate=cn,
        get_word_completions=gwc,
        choose_words=cw,
    )
    # this needs a real reactor, for blockingCallFromThread
    c = CodeInputter(helper, reactor)
    cabc = c._commit_and_build_completions

    # in this test, we pretend that nameplates 1,12,34 are active.

    # 43 TAB -> nothing (and refresh_nameplates)
    gnc.configure_mock(return_value=[])
    matches = await deferToThread(cabc, "43")
    assert matches == []
    assert rn.mock_calls == [mock.call()]
    assert gnc.mock_calls == [mock.call("43")]
    assert cn.mock_calls == []
    rn.reset_mock()
    gnc.reset_mock()

    # 1 TAB -> 1-, 12- (and refresh_nameplates)
    gnc.configure_mock(return_value=["1-", "12-"])
    matches = await deferToThread(cabc, "1")
    assert matches == ["1-", "12-"]
    assert rn.mock_calls == [mock.call()]
    assert gnc.mock_calls == [mock.call("1")]
    assert cn.mock_calls == []
    rn.reset_mock()
    gnc.reset_mock()

    # 12 TAB -> 12- (and refresh_nameplates)
    # I wouldn't mind if it didn't refresh the nameplates here, but meh
    gnc.configure_mock(return_value=["12-"])
    matches = await deferToThread(cabc, "12")
    assert matches == ["12-"]
    assert rn.mock_calls == [mock.call()]
    assert gnc.mock_calls == [mock.call("12")]
    assert cn.mock_calls == []
    rn.reset_mock()
    gnc.reset_mock()

    # 12- TAB -> 12- {all words} (claim, no refresh)
    gnc.configure_mock(return_value=["12-"])
    gwc.configure_mock(return_value=["and-", "ark-", "aaah!zombies!!-"])
    matches = await deferToThread(cabc, "12-")
    assert matches == ["12-aaah!zombies!!-", "12-and-", "12-ark-"]
    assert rn.mock_calls == []
    assert gnc.mock_calls == []
    assert cn.mock_calls == [mock.call("12")]
    assert gwc.mock_calls == [mock.call("")]
    cn.reset_mock()
    gwc.reset_mock()

    # TODO: another path with "3 TAB" then "34-an TAB", so the claim
    # happens in the second call (and it waits for the wordlist)

    # 12-a TAB -> 12-and- 12-ark- 12-aaah!zombies!!-
    gnc.configure_mock(side_effect=ValueError)
    gwc.configure_mock(return_value=["and-", "ark-", "aaah!zombies!!-"])
    matches = await deferToThread(cabc, "12-a")
    # matches are always sorted
    assert matches == ["12-aaah!zombies!!-", "12-and-", "12-ark-"]
    assert rn.mock_calls == []
    assert gnc.mock_calls == []
    assert cn.mock_calls == []
    assert gwc.mock_calls == [mock.call("a")]
    gwc.reset_mock()

    # 12-and-b TAB -> 12-and-bat 12-and-bet 12-and-but
    gnc.configure_mock(side_effect=ValueError)
    # wordlist knows the code length, so doesn't add hyphens to these
    gwc.configure_mock(return_value=["and-bat", "and-bet", "and-but"])
    matches = await deferToThread(cabc, "12-and-b")
    assert matches == ["12-and-bat", "12-and-bet", "12-and-but"]
    assert rn.mock_calls == []
    assert gnc.mock_calls == []
    assert cn.mock_calls == []
    assert gwc.mock_calls == [mock.call("and-b")]
    gwc.reset_mock()

    await deferToThread(c.finish, "12-and-bat")
    assert cw.mock_calls == [mock.call("and-bat")]

def test_incomplete_code():
    helper = mock.Mock()
    c = CodeInputter(helper, "reactor")
    c.bcft = fake_blockingCallFromThread
    with pytest.raises(KeyFormatError) as e:
        c.finish("1")
    assert str(e.value) == "incomplete wormhole code"

@ensureDeferred
async def test_rollback_nameplate_during_completion():
    helper = mock.Mock()
    gwc = helper.get_word_completions = mock.Mock()
    gwc.configure_mock(return_value=["code", "court"])
    c = CodeInputter(helper, reactor)
    cabc = c._commit_and_build_completions
    matches = await deferToThread(cabc, "1-co")  # this commits us to 1-
    assert helper.mock_calls == [
        mock.call.choose_nameplate("1"),
        mock.call.when_wordlist_is_available(),
        mock.call.get_word_completions("co")
    ]
    assert matches == ["1-code", "1-court"]
    helper.reset_mock()
    with pytest.raises(AlreadyInputNameplateError) as e:
        await deferToThread(cabc, "2-co")
    assert str(e.value) == "nameplate (1-) already entered, cannot go back"
    assert helper.mock_calls == []

@ensureDeferred
async def test_rollback_nameplate_during_finish():
    helper = mock.Mock()
    gwc = helper.get_word_completions = mock.Mock()
    gwc.configure_mock(return_value=["code", "court"])
    c = CodeInputter(helper, reactor)
    cabc = c._commit_and_build_completions
    matches = await deferToThread(cabc, "1-co")  # this commits us to 1-
    assert helper.mock_calls == [
        mock.call.choose_nameplate("1"),
        mock.call.when_wordlist_is_available(),
        mock.call.get_word_completions("co")
    ]
    assert matches == ["1-code", "1-court"]
    helper.reset_mock()
    with pytest.raises(AlreadyInputNameplateError) as e:
        await deferToThread(c.finish, "2-code")
    assert str(e.value) == "nameplate (1-) already entered, cannot go back"
    assert helper.mock_calls == []

@mock.patch("wormhole._rlcompleter.stderr")
def test_warn_readline(stderr):
    # there is no good way to test that this function gets used at the
    # right time, since it involves a reactor and a "system event
    # trigger", but let's at least make sure it's invocable
    warn_readline()
    expected = "\nCommand interrupted: please press Return to quit"
    assert stderr.mock_calls == \
                     [mock.call.write(expected),
                      mock.call.write("\n")]



from attr import attrib, attrs
from attr.validators import instance_of
from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces
from ._key import derive_phase_key, encrypt_data
from .util import provides


@attrs
@implementer(_interfaces.ISend)
class Send:
    _side = attrib(validator=instance_of(str))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._queue = []

    def wire(self, mailbox):
        self._M = _interfaces.IMailbox(mailbox)

    @m.state(initial=True)
    def S0_no_key(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def S1_verified_key(self):
        pass  # pragma: no cover

    # from Receive
    @m.input()
    def got_verified_key(self, key):
        pass

    # from Boss
    @m.input()
    def send(self, phase, plaintext):
        pass

    @m.output()
    def queue(self, phase, plaintext):
        assert isinstance(phase, str), type(phase)
        assert isinstance(plaintext, bytes), type(plaintext)
        self._queue.append((phase, plaintext))

    @m.output()
    def record_key(self, key):
        self._key = key

    @m.output()
    def drain(self, key):
        del key
        for (phase, plaintext) in self._queue:
            self._encrypt_and_send(phase, plaintext)
        self._queue[:] = []

    @m.output()
    def deliver(self, phase, plaintext):
        assert isinstance(phase, str), type(phase)
        assert isinstance(plaintext, bytes), type(plaintext)
        self._encrypt_and_send(phase, plaintext)

    def _encrypt_and_send(self, phase, plaintext):
        assert self._key
        data_key = derive_phase_key(self._key, self._side, phase)
        encrypted = encrypt_data(data_key, plaintext)
        self._M.add_message(phase, encrypted)

    S0_no_key.upon(send, enter=S0_no_key, outputs=[queue])
    S0_no_key.upon(
        got_verified_key, enter=S1_verified_key, outputs=[record_key, drain])
    S1_verified_key.upon(send, enter=S1_verified_key, outputs=[deliver])



import os
from os.path import exists, expanduser, join

import click
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

from .. import xfer_util


class PubkeyError(Exception):
    pass


def find_public_key(hint=None):
    """
    This looks for an appropriate SSH key to send, possibly querying
    the user in the meantime. DO NOT CALL after reactor.run as this
    (possibly) does blocking stuff like asking the user questions (via
    click.prompt())

    Returns a 3-tuple: kind, keyid, pubkey_data
    """

    if hint is None:
        hint = expanduser('~/.ssh/')
    else:
        if not exists(hint):
            raise PubkeyError(f"Can't find '{hint}'")

    pubkeys = [f for f in os.listdir(hint) if f.endswith('.pub')]
    if len(pubkeys) == 0:
        raise PubkeyError(f"No public keys in '{hint}'")
    elif len(pubkeys) > 1:
        got_key = False
        while not got_key:
            ans = click.prompt(
                "Multiple public-keys found:\n" +
                "\n".join([f"  {a}: {b}"
                           for a, b in enumerate(pubkeys)]) +
                "\nSend which one?"
            )
            try:
                ans = int(ans)
                if ans < 0 or ans >= len(pubkeys):
                    ans = None
                else:
                    got_key = True
                    with open(join(hint, pubkeys[ans])) as f:
                        pubkey = f.read()

            except Exception:
                got_key = False
    else:
        with open(join(hint, pubkeys[0])) as f:
            pubkey = f.read()
    parts = pubkey.strip().split(maxsplit=2)
    kind = parts[0]
    keyid = 'unknown' if len(parts) <= 2 else parts[2]

    return kind, keyid, pubkey


@inlineCallbacks
def accept(cfg, reactor=reactor):
    yield xfer_util.send(
        reactor,
        cfg.appid or "lothar.com/wormhole/ssh-add",
        cfg.relay_url,
        data=cfg.public_key[2],
        code=cfg.code,
        use_tor=cfg.tor,
        launch_tor=cfg.launch_tor,
        tor_control_port=cfg.tor_control_port,
    )
    print("Key sent.")


@inlineCallbacks
def invite(cfg, reactor=reactor):
    def on_code_created(code):
        print("Now tell the other user to run:")
        print()
        print(f"wormhole ssh accept {code}")
        print()

    if cfg.ssh_user is None:
        ssh_path = expanduser('~/.ssh/')
    else:
        ssh_path = expanduser(f'~{cfg.ssh_user}/.ssh/')
    auth_key_path = join(ssh_path, 'authorized_keys')
    if not exists(auth_key_path):
        print(f"Note: '{auth_key_path}' not found; will be created")
        if not exists(ssh_path):
            print(f"      '{ssh_path}' doesn't exist either")
    else:
        try:
            open(auth_key_path, 'a').close()
        except OSError:
            print(f"No write permission on '{auth_key_path}'")
            return
    try:
        os.listdir(ssh_path)
    except OSError:
        print(f"Can't read '{ssh_path}'")
        return

    pubkey = yield xfer_util.receive(
        reactor,
        cfg.appid or "lothar.com/wormhole/ssh-add",
        cfg.relay_url,
        None,  # allocate a code for us
        use_tor=cfg.tor,
        launch_tor=cfg.launch_tor,
        tor_control_port=cfg.tor_control_port,
        on_code=on_code_created,
    )

    parts = pubkey.split()
    kind = parts[0]
    keyid = 'unknown' if len(parts) <= 2 else parts[2]

    if not exists(auth_key_path):
        if not exists(ssh_path):
            os.mkdir(ssh_path, mode=0o700)
    with open(auth_key_path, 'a', 0o600) as f:
        f.write(f'{pubkey.strip()}\n')
    print(f"Appended key type='{kind}' id='{keyid}' to '{auth_key_path}'")



from twisted.internet.defer import ensureDeferred
from twisted.python import log

import pytest
import pytest_twisted

from .common import setup_mailbox, setup_transit_relay

# from kyle altendorf
# see also https://github.com/pytest-dev/pytest-twisted/issues/4
import gc


@pytest.fixture(scope="session")
def reactor():
    from twisted.internet import reactor
    yield reactor


@pytest.fixture(scope="session")
def mailbox(reactor):
    from wormhole import __version__
    mb = pytest_twisted.blockon(
        ensureDeferred(
            setup_mailbox(reactor, advertise_version=str(__version__))
        )
    )
    mb.service.startService()
    yield mb
    pytest_twisted.blockon(mb.service.stopService())
    pytest_twisted.blockon(mb.port.stopListening())
    ##from twisted.internet import task
    ##pytest_twisted.blockon(task.deferLater(reactor, 0.1, lambda: None))


@pytest.fixture(scope="session")
def transit_relay(reactor):
    url, service = setup_transit_relay(reactor)
    service.startService()
    yield url
    pytest_twisted.blockon(service.stopService())





class Observer:
    def __init__(self):
        self.failures = []

    def __call__(self, event_dict):
        is_error = event_dict.get('isError')
        if is_error:
            self.failures.append(event_dict["failure"])

    def flush(self, klass):
        flushed = [
            f
            for f in self.failures
            if isinstance(f.value, klass)
        ]
        self.failures = [
            f
            for f in self.failures
            if not isinstance(f.value, klass)
        ]
        return flushed

    def assert_empty(self):
        assert [] == self.failures


@pytest.fixture
def observe_errors():
    observer = Observer()
    gc.collect()
    log.startLoggingWithObserver(observer, 0)

    yield observer

    gc.collect()
    log.removeObserver(observer)
    observer.assert_empty()



import io

from twisted.internet import defer
from twisted.internet.error import ConnectError

from unittest import mock
import pytest
import pytest_twisted

from .._interfaces import ITorManager
from ..errors import NoTorError
from ..tor_manager import SocksOnlyTor, get_tor


class X():
    pass


@pytest_twisted.ensureDeferred
async def test_no_txtorcon():
    with mock.patch("wormhole.tor_manager.txtorcon", None):
        with pytest.raises(NoTorError):
            await get_tor(None)


@pytest_twisted.ensureDeferred
async def test_bad_args():
    with pytest.raises(TypeError) as f:
        await get_tor(None, launch_tor="not boolean")
    assert str(f.value) == "launch_tor= must be boolean"

    with pytest.raises(TypeError) as f:
        await get_tor(None, tor_control_port=1234)
    assert str(f.value) == "tor_control_port= must be str or None"

    with pytest.raises(ValueError) as f:
        await get_tor(None, launch_tor=True, tor_control_port="tcp:127.0.0.1:1234")
    assert str(f.value) == "cannot combine --launch-tor and --tor-control-port="


@pytest_twisted.ensureDeferred
async def test_launch():
    reactor = object()
    my_tor = X()  # object() didn't like providedBy()
    launch_d = defer.Deferred()
    stderr = io.StringIO()
    with mock.patch(
            "wormhole.tor_manager.txtorcon.launch",
            side_effect=launch_d) as launch:
        d = get_tor(reactor, launch_tor=True, stderr=stderr)
        assert not d.called
        assert launch.mock_calls == [mock.call(reactor)]
        launch_d.callback(my_tor)
        tor = await d
        assert tor is my_tor
        assert ITorManager.providedBy(tor)
        assert stderr.getvalue() == \
            " launching a new Tor process, this may take a while..\n"


@pytest_twisted.ensureDeferred
async def test_connect():
    reactor = object()
    my_tor = X()  # object() didn't like providedBy()
    connect_d = defer.Deferred()
    stderr = io.StringIO()
    with mock.patch(
            "wormhole.tor_manager.txtorcon.connect",
            side_effect=connect_d) as connect:
        with mock.patch(
                "wormhole.tor_manager.clientFromString",
                side_effect=["foo"]) as sfs:
            d = get_tor(reactor, stderr=stderr)
    assert sfs.mock_calls == []
    assert not d.called
    assert connect.mock_calls == [mock.call(reactor)]
    connect_d.callback(my_tor)
    tor = await d
    assert tor is my_tor
    assert ITorManager.providedBy(tor)
    assert stderr.getvalue() == \
                     " using Tor via default control port\n"


@pytest_twisted.ensureDeferred
async def test_connect_fails():
    reactor = object()
    connect_d = defer.Deferred()
    stderr = io.StringIO()
    with mock.patch(
            "wormhole.tor_manager.txtorcon.connect",
            side_effect=connect_d) as connect:
        with mock.patch(
                "wormhole.tor_manager.clientFromString",
                side_effect=["foo"]) as sfs:
            d = get_tor(reactor, stderr=stderr)
    assert sfs.mock_calls == []
    assert not d.called
    assert connect.mock_calls == [mock.call(reactor)]

    connect_d.errback(ConnectError())
    tor = await d
    assert isinstance(tor, SocksOnlyTor)
    assert ITorManager.providedBy(tor)
    assert tor._reactor == reactor
    assert stderr.getvalue() == \
        " unable to find default Tor control port, using SOCKS\n"


@pytest_twisted.ensureDeferred
async def test_connect_custom_control_port():
    reactor = object()
    my_tor = X()  # object() didn't like providedBy()
    tcp = "PORT"
    ep = object()
    connect_d = defer.Deferred()
    stderr = io.StringIO()
    with mock.patch(
            "wormhole.tor_manager.txtorcon.connect",
            side_effect=connect_d) as connect:
        with mock.patch(
                "wormhole.tor_manager.clientFromString",
                side_effect=[ep]) as sfs:
            d = get_tor(reactor, tor_control_port=tcp, stderr=stderr)
    assert sfs.mock_calls == [mock.call(reactor, tcp)]
    assert not d.called
    assert connect.mock_calls == [mock.call(reactor, ep)]
    connect_d.callback(my_tor)
    tor = await d
    assert tor is my_tor
    assert ITorManager.providedBy(tor)
    assert stderr.getvalue() == \
                     " using Tor via control port at PORT\n"


@pytest_twisted.ensureDeferred
async def test_connect_custom_control_port_fails():
    reactor = object()
    tcp = "port"
    ep = object()
    connect_d = defer.Deferred()
    stderr = io.StringIO()
    with mock.patch(
            "wormhole.tor_manager.txtorcon.connect",
            side_effect=connect_d) as connect:
        with mock.patch(
                "wormhole.tor_manager.clientFromString",
                side_effect=[ep]) as sfs:
            d = get_tor(reactor, tor_control_port=tcp, stderr=stderr)
    assert sfs.mock_calls == [mock.call(reactor, tcp)]
    assert not d.called
    assert connect.mock_calls == [mock.call(reactor, ep)]

    connect_d.errback(ConnectError())
    with pytest.raises(ConnectError):
        await d
    assert stderr.getvalue() == ""


def test_tor():
    reactor = object()
    sot = SocksOnlyTor(reactor)
    fake_ep = object()
    with mock.patch(
            "wormhole.tor_manager.txtorcon.TorClientEndpoint",
            return_value=fake_ep) as tce:
        ep = sot.stream_via("host", "port")
    assert ep is fake_ep
    assert tce.mock_calls == [
        mock.call(
            "host",
            "port",
            socks_endpoint=None,
            tls=False,
            reactor=reactor)
    ]



def handle_welcome(welcome, relay_url, my_version, stderr):
    if "motd" in welcome:
        motd_lines = welcome["motd"].splitlines()
        motd_formatted = "\n ".join(motd_lines)
        print(
            f"Server (at {relay_url}) says:\n {motd_formatted}",
            file=stderr)

    # Only warn if we're running a release version (e.g. 0.0.6, not
    # 0.0.6+DISTANCE.gHASH). Only warn once.
    if ("current_cli_version" in welcome and
         "+" not in my_version and
         welcome["current_cli_version"] != my_version):
        print(
            ("Warning: errors may occur unless both sides are running the"
             " same version"),
            file=stderr)
        print(
            "Server claims %s is current, but ours is %s" %
            (welcome["current_cli_version"], my_version),
            file=stderr)



from attr import attrib, attrs
from attr.validators import instance_of
from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces
from .util import provides


@attrs
@implementer(_interfaces.IOrder)
class Order:
    _side = attrib(validator=instance_of(str))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._key = None
        self._queue = []

    def wire(self, key, receive):
        self._K = _interfaces.IKey(key)
        self._R = _interfaces.IReceive(receive)

    @m.state(initial=True)
    def S0_no_pake(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def S1_yes_pake(self):
        pass  # pragma: no cover

    def got_message(self, side, phase, body):
        # print("ORDER[%s].got_message(%s)" % (self._side, phase))
        assert isinstance(side, str), type(phase)
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), type(body)
        if phase == "pake":
            self.got_pake(side, phase, body)
        else:
            self.got_non_pake(side, phase, body)

    @m.input()
    def got_pake(self, side, phase, body):
        pass

    @m.input()
    def got_non_pake(self, side, phase, body):
        pass

    @m.output()
    def queue(self, side, phase, body):
        assert isinstance(side, str), type(phase)
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), type(body)
        self._queue.append((side, phase, body))

    @m.output()
    def notify_key(self, side, phase, body):
        self._K.got_pake(body)

    @m.output()
    def drain(self, side, phase, body):
        del phase
        del body
        for (side, phase, body) in self._queue:
            self._deliver(side, phase, body)
        self._queue[:] = []

    @m.output()
    def deliver(self, side, phase, body):
        self._deliver(side, phase, body)

    def _deliver(self, side, phase, body):
        self._R.got_message(side, phase, body)

    S0_no_pake.upon(got_non_pake, enter=S0_no_pake, outputs=[queue])
    S0_no_pake.upon(got_pake, enter=S1_yes_pake, outputs=[notify_key, drain])
    S1_yes_pake.upon(got_non_pake, enter=S1_yes_pake, outputs=[deliver])



import os
import socket
import sys
import time
from binascii import hexlify, unhexlify
from collections import deque

from nacl.secret import SecretBox
from twisted.internet import (address, defer, endpoints, error, interfaces,
                              protocol, task)
from twisted.internet.defer import inlineCallbacks
from twisted.protocols import policies
from twisted.python import log
from twisted.python.runtime import platformType
from zope.interface import implementer

from . import ipaddrs
from .errors import InternalError
from .timing import DebugTiming
from .util import bytes_to_hexstr, HKDF
from ._hints import (DirectTCPV1Hint, RelayV1Hint,
                     parse_hint_argv, describe_hint_obj, endpoint_from_hint_obj,
                     parse_tcp_v1_hint)


class TransitError(Exception):
    pass


class BadHandshake(Exception):
    pass


class TransitClosed(TransitError):
    pass


class BadNonce(TransitError):
    pass


# The beginning of each TCP connection consists of the following handshake
# messages. The sender transmits the same text regardless of whether it is on
# the initiating/connecting end of the TCP connection, or on the
# listening/accepting side. Same for the receiver.
#
#  sender -> receiver: transit sender TXID_HEX ready\n\n
#  receiver -> sender: transit receiver RXID_HEX ready\n\n
#
# Any deviations from this result in the socket being closed. The handshake
# messages are designed to provoke an invalid response from other sorts of
# servers (HTTP, SMTP, echo).
#
# If the sender is satisfied with the handshake, and this is the first socket
# to complete negotiation, the sender does:
#
#  sender -> receiver: go\n
#
# and the next byte on the wire will be from the application.
#
# If this is not the first socket, the sender does:
#
#  sender -> receiver: nevermind\n
#
# and closes the socket.

# So the receiver looks for "transit sender TXID_HEX ready\n\ngo\n" and hangs
# up upon the first wrong byte. The sender lookgs for "transit receiver
# RXID_HEX ready\n\n" and then makes a first/not-first decision about sending
# "go\n" or "nevermind\n"+close().


def build_receiver_handshake(key):
    hexid = HKDF(key, 32, CTXinfo=b"transit_receiver")
    return b"transit receiver " + hexlify(hexid) + b" ready\n\n"


def build_sender_handshake(key):
    hexid = HKDF(key, 32, CTXinfo=b"transit_sender")
    return b"transit sender " + hexlify(hexid) + b" ready\n\n"


def build_sided_relay_handshake(key, side):
    assert isinstance(side, str)
    assert len(side) == 8 * 2
    token = HKDF(key, 32, CTXinfo=b"transit_relay_token")
    return b"please relay " + hexlify(token) + b" for side " + side.encode(
        "ascii") + b"\n"


TIMEOUT = 60  # seconds


@implementer(interfaces.IProducer, interfaces.IConsumer)
class Connection(protocol.Protocol, policies.TimeoutMixin):
    def __init__(self, owner, relay_handshake, start, description):
        self.state = "too-early"
        self.buf = b""
        self.owner = owner
        self.relay_handshake = relay_handshake
        self.start = start
        self._description = description
        self._negotiation_d = defer.Deferred(self._cancel)
        self._error = None
        self._consumer = None
        self._consumer_bytes_written = 0
        self._consumer_bytes_expected = None
        self._consumer_deferred = None
        self._inbound_records = deque()
        self._waiting_reads = deque()

    def connectionMade(self):
        self.setTimeout(TIMEOUT)  # does timeoutConnection() when it expires
        self.factory.connectionWasMade(self)

    def startNegotiation(self):
        if self.relay_handshake is not None:
            self.transport.write(self.relay_handshake)
            self.state = "relay"
        else:
            self.state = "start"
        self.dataReceived(b"")  # cycle the state machine
        return self._negotiation_d

    def _cancel(self, d):
        self.state = "hung up"  # stop reacting to anything further
        self._error = defer.CancelledError()
        self.transport.loseConnection()
        # if connectionLost isn't called synchronously, then our
        # self._negotiation_d will have been errbacked by Deferred.cancel
        # (which is our caller). So if it's still around, clobber it
        if self._negotiation_d:
            self._negotiation_d = None

    def dataReceived(self, data):
        try:
            self._dataReceived(data)
        except Exception as e:
            self.setTimeout(None)
            self._error = e
            self.transport.loseConnection()
            self.state = "hung up"
            if not isinstance(e, BadHandshake):
                raise

    def _check_and_remove(self, expected):
        # any divergence is a handshake error
        if not self.buf.startswith(expected[:len(self.buf)]):
            raise BadHandshake(f"got {self.buf!r} want {expected!r}")
        if len(self.buf) < len(expected):
            return False  # keep waiting
        self.buf = self.buf[len(expected):]
        return True

    def _dataReceived(self, data):
        # protocol is:
        #  (maybe: send relay handshake, wait for ok)
        #  send (send|receive)_handshake
        #  wait for (receive|send)_handshake
        #  sender: decide, send "go" or hang up
        #  receiver: wait for "go"
        self.buf += data

        assert self.state != "too-early"
        if self.state == "relay":
            if not self._check_and_remove(b"ok\n"):
                return
            self.state = "start"
        if self.state == "start":
            self.transport.write(self.owner._send_this())
            self.state = "handshake"
        if self.state == "handshake":
            if not self._check_and_remove(self.owner._expect_this()):
                return
            self.state = self.owner.connection_ready(self)
            # If we're the receiver, we'll be moved to state
            # "wait-for-decision", which means we're waiting for the other
            # side (the sender) to make a decision. If we're the sender,
            # we'll either be moved to state "go" (send GO and move directly
            # to state "records") or state "nevermind" (send NEVERMIND and
            # hang up).

        if self.state == "wait-for-decision":
            if not self._check_and_remove(b"go\n"):
                return
            self._negotiationSuccessful()
        if self.state == "go":
            GO = b"go\n"
            self.transport.write(GO)
            self._negotiationSuccessful()
        if self.state == "nevermind":
            self.transport.write(b"nevermind\n")
            raise BadHandshake("abandoned")
        if self.state == "records":
            return self.dataReceivedRECORDS()
        if self.state == "hung up":
            return
        if isinstance(self.state, Exception):  # for tests
            raise self.state
        raise ValueError(f"internal error: unknown state {self.state}")

    def _negotiationSuccessful(self):
        self.state = "records"
        self.setTimeout(None)
        send_key = self.owner._sender_record_key()
        self.send_box = SecretBox(send_key)
        self.send_nonce = 0
        receive_key = self.owner._receiver_record_key()
        self.receive_box = SecretBox(receive_key)
        self.next_receive_nonce = 0
        d, self._negotiation_d = self._negotiation_d, None
        d.callback(self)

    def dataReceivedRECORDS(self):
        while True:
            if len(self.buf) < 4:
                return
            length = int(hexlify(self.buf[:4]), 16)
            if len(self.buf) < 4 + length:
                return
            encrypted, self.buf = self.buf[4:4 + length], self.buf[4 + length:]

            record = self._decrypt_record(encrypted)
            self.recordReceived(record)

    def _decrypt_record(self, encrypted):
        nonce_buf = encrypted[:SecretBox.NONCE_SIZE]  # assume it's prepended
        nonce = int(hexlify(nonce_buf), 16)
        if nonce != self.next_receive_nonce:
            raise BadNonce(
                "received out-of-order record: got %d, expected %d" %
                (nonce, self.next_receive_nonce))
        self.next_receive_nonce += 1
        record = self.receive_box.decrypt(encrypted)
        return record

    def describe(self):
        return self._description

    def send_record(self, record):
        if not isinstance(record, bytes):
            raise InternalError
        assert SecretBox.NONCE_SIZE == 24
        assert self.send_nonce < 2**(8 * 24)
        assert len(record) < 2**(8 * 4)
        nonce = unhexlify(f"{self.send_nonce:048x}")  # big-endian
        self.send_nonce += 1
        encrypted = self.send_box.encrypt(record, nonce)
        length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
        self.transport.write(length)
        self.transport.write(encrypted)

    def recordReceived(self, record):
        if self._consumer:
            self._writeToConsumer(record)
            return
        self._inbound_records.append(record)
        self._deliverRecords()

    def receive_record(self):
        d = defer.Deferred()
        self._waiting_reads.append(d)
        self._deliverRecords()
        return d

    def _deliverRecords(self):
        while self._inbound_records and self._waiting_reads:
            r = self._inbound_records.popleft()
            d = self._waiting_reads.popleft()
            d.callback(r)

    def close(self):
        self.transport.loseConnection()
        while self._waiting_reads:
            d = self._waiting_reads.popleft()
            d.errback(error.ConnectionClosed())

    def timeoutConnection(self):
        self._error = BadHandshake("timeout")
        self.transport.loseConnection()

    def connectionLost(self, reason=None):
        self.setTimeout(None)
        while self._waiting_reads:
            d = self._waiting_reads.popleft()
            d.errback(error.ConnectionClosed())

        d, self._negotiation_d = self._negotiation_d, None
        # the Deferred is only relevant until negotiation finishes, so skip
        # this if it's already been fired
        if d:
            # Each call to loseConnection() sets self._error first, so we can
            # deliver useful information to the Factory that's waiting on
            # this (although they'll generally ignore the specific error,
            # except for logging unexpected ones). The possible cases are:
            #
            # cancel: defer.CancelledError
            # far-end disconnect: BadHandshake("connection lost")
            # handshake error (something we didn't like): BadHandshake(what)
            # other error: some other Exception
            # timeout: BadHandshake("timeout")

            d.errback(self._error or BadHandshake("connection lost"))
        if self._consumer_deferred:
            self._consumer_deferred.errback(error.ConnectionClosed())

    # IConsumer methods, for outbound flow-control. We pass these through to
    # the transport. The 'producer' is something like a t.p.basic.FileSender
    def registerProducer(self, producer, streaming):
        assert interfaces.IConsumer.providedBy(self.transport)
        self.transport.registerProducer(producer, streaming)

    def unregisterProducer(self):
        self.transport.unregisterProducer()

    def write(self, data):
        self.send_record(data)

    # IProducer methods, for inbound flow-control. We pass these through to
    # the transport.
    def stopProducing(self):
        self.transport.stopProducing()

    def pauseProducing(self):
        self.transport.pauseProducing()

    def resumeProducing(self):
        self.transport.resumeProducing()

    # Helper methods

    def connectConsumer(self, consumer, expected=None):
        """Helper method to glue an instance of e.g. t.p.ftp.FileConsumer to
        us. Inbound records will be written as bytes to the consumer.

        Set 'expected' to an integer to automatically disconnect when at
        least that number of bytes have been written. This function will then
        return a Deferred (that fires with the number of bytes actually
        received). If the connection is lost while this Deferred is
        outstanding, it will errback. If 'expected' is 0, the Deferred will
        fire right away.

        If 'expected' is None, then this function returns None instead of a
        Deferred, and you must call disconnectConsumer() when you are done."""

        if self._consumer:
            raise RuntimeError(
                f"A consumer is already attached: {self._consumer!r}")

        # be aware of an ordering hazard: when we call the consumer's
        # .registerProducer method, they are likely to immediately call
        # self.resumeProducing, which we'll deliver to self.transport, which
        # might call our .dataReceived, which may cause more records to be
        # available. By waiting to set self._consumer until *after* we drain
        # any pending records, we avoid delivering records out of order,
        # which would be bad.
        consumer.registerProducer(self, True)
        # There might be enough data queued to exceed 'expected' before we
        # leave this function. We must be sure to register the producer
        # before it gets unregistered.

        self._consumer = consumer
        self._consumer_bytes_written = 0
        self._consumer_bytes_expected = expected
        d = None
        if expected is not None:
            d = defer.Deferred()
        self._consumer_deferred = d
        if expected == 0:
            # write empty record to kick consumer into shutdown
            self._writeToConsumer(b"")
        # drain any pending records
        while self._consumer and self._inbound_records:
            r = self._inbound_records.popleft()
            self._writeToConsumer(r)
        return d

    def _writeToConsumer(self, record):
        self._consumer.write(record)
        self._consumer_bytes_written += len(record)
        if self._consumer_bytes_expected is not None:
            if self._consumer_bytes_written >= self._consumer_bytes_expected:
                d = self._consumer_deferred
                self.disconnectConsumer()
                d.callback(self._consumer_bytes_written)

    def disconnectConsumer(self):
        self._consumer.unregisterProducer()
        self._consumer = None
        self._consumer_bytes_expected = None
        self._consumer_deferred = None

    # Helper method to write a known number of bytes to a file. This has no
    # flow control: the filehandle cannot push back. 'progress' is an
    # optional callable which will be called on each write (with the number
    # of bytes written). Returns a Deferred that fires (with the number of
    # bytes written) when the count is reached or the RecordPipe is closed.

    def writeToFile(self, f, expected, progress=None, hasher=None):
        fc = FileConsumer(f, progress, hasher)
        return self.connectConsumer(fc, expected)


class OutboundConnectionFactory(protocol.ClientFactory):
    protocol = Connection

    def __init__(self, owner, relay_handshake, description):
        self.owner = owner
        self.relay_handshake = relay_handshake
        self._description = description
        self.start = time.time()

    def buildProtocol(self, addr):
        p = self.protocol(self.owner, self.relay_handshake, self.start,
                          self._description)
        p.factory = self
        return p

    def connectionWasMade(self, p):
        # outbound connections are handled via the endpoint
        pass


class InboundConnectionFactory(protocol.ClientFactory):
    protocol = Connection

    def __init__(self, owner):
        self.owner = owner
        self.start = time.time()
        self._inbound_d = defer.Deferred(self._cancel)
        self._pending_connections = set()

    def whenDone(self):
        return self._inbound_d

    def _cancel(self, inbound_d):
        self._shutdown()
        # our _inbound_d will be errbacked by Deferred.cancel()

    def _shutdown(self):
        for d in list(self._pending_connections):
            d.cancel()  # that fires _remove and _proto_failed

    def _describePeer(self, addr):
        if isinstance(addr, address.HostnameAddress):
            return "<-%s:%d" % (addr.hostname, addr.port)
        elif isinstance(addr, (address.IPv4Address, address.IPv6Address)):
            return "<-%s:%d" % (addr.host, addr.port)
        return f"<-{addr!r}"

    def buildProtocol(self, addr):
        p = self.protocol(self.owner, None, self.start,
                          self._describePeer(addr))
        p.factory = self
        return p

    def connectionWasMade(self, p):
        d = p.startNegotiation()
        self._pending_connections.add(d)
        d.addBoth(self._remove, d)
        d.addCallbacks(self._proto_succeeded, self._proto_failed)

    def _remove(self, res, d):
        self._pending_connections.remove(d)
        return res

    def _proto_succeeded(self, p):
        self._shutdown()
        self._inbound_d.callback(p)

    def _proto_failed(self, f):
        # ignore these two, let Twisted log everything else
        f.trap(BadHandshake, defer.CancelledError)


def allocate_tcp_port():
    """Return an (integer) available TCP port on localhost. This briefly
    listens on the port in question, then closes it right away."""
    # We want to bind() the socket but not listen(). Twisted (in
    # tcp.Port.createInternetSocket) would do several other things:
    # non-blocking, close-on-exec, and SO_REUSEADDR. We don't need
    # non-blocking because we never listen on it, and we don't need
    # close-on-exec because we close it right away. So just add SO_REUSEADDR.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if platformType == "posix" and sys.platform != "cygwin":
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(("127.0.0.1", 0))
    port = s.getsockname()[1]
    s.close()
    return port


class _ThereCanBeOnlyOne:
    """Accept a list of contender Deferreds, and return a summary Deferred.
    When the first contender fires successfully, cancel the rest and fire the
    summary with the winning contender's result. If all error, errback the
    summary.

    status_cb=?
    """

    def __init__(self, contenders):
        self._remaining = set(contenders)
        self._winner_d = defer.Deferred(self._cancel)
        self._first_success = None
        self._first_failure = None
        self._have_winner = False
        self._fired = False

    def _cancel(self, _):
        for d in list(self._remaining):
            d.cancel()
        # since that will errback everything in _remaining, we'll have hit
        # _maybe_done() and fired self._winner_d by this point

    def run(self):
        for d in list(self._remaining):
            d.addBoth(self._remove, d)
            d.addCallbacks(self._succeeded, self._failed)
            d.addCallback(self._maybe_done)
        return self._winner_d

    def _remove(self, res, d):
        self._remaining.remove(d)
        return res

    def _succeeded(self, res):
        self._have_winner = True
        self._first_success = res
        for d in list(self._remaining):
            d.cancel()

    def _failed(self, f):
        if self._first_failure is None:
            self._first_failure = f

    def _maybe_done(self, _):
        if self._remaining:
            return
        if self._fired:
            return
        self._fired = True
        if self._have_winner:
            self._winner_d.callback(self._first_success)
        else:
            self._winner_d.errback(self._first_failure)


def there_can_be_only_one(contenders):
    return _ThereCanBeOnlyOne(contenders).run()


class Common:
    RELAY_DELAY = 2.0
    TRANSIT_KEY_LENGTH = SecretBox.KEY_SIZE

    def __init__(self,
                 transit_relay,
                 no_listen=False,
                 tor=None,
                 reactor=None,
                 timing=None):
        self._side = bytes_to_hexstr(os.urandom(8))  # unicode
        if transit_relay:
            if not isinstance(transit_relay, str):
                raise InternalError
            # TODO: allow multiple hints for a single relay
            relay_hint = parse_hint_argv(transit_relay)
            relay = RelayV1Hint(hints=(relay_hint, ))
            self._transit_relays = [relay]
        else:
            self._transit_relays = []
        self._their_direct_hints = []  # hintobjs
        self._our_relay_hints = set(self._transit_relays)
        self._tor = tor
        self._transit_key = None
        self._no_listen = no_listen
        self._waiting_for_transit_key = []
        self._listener = None
        self._winner = None
        if reactor is None:
            from twisted.internet import reactor
        self._reactor = reactor
        self._timing = timing or DebugTiming()
        self._timing.add("transit")

    def _build_listener(self):
        if self._no_listen or self._tor:
            return ([], None)
        portnum = allocate_tcp_port()
        addresses = ipaddrs.find_addresses()
        non_loopback_addresses = [a for a in addresses if a != "127.0.0.1"]
        if non_loopback_addresses:
            # some test hosts, including the appveyor VMs, *only* have
            # 127.0.0.1, and the tests will hang badly if we remove it.
            addresses = non_loopback_addresses
        direct_hints = [
            DirectTCPV1Hint(str(addr), portnum, 0.0) for addr in addresses
        ]
        ep = endpoints.serverFromString(self._reactor, "tcp:%d" % portnum)
        return direct_hints, ep

    def get_connection_abilities(self):
        return [
            {
                "type": "direct-tcp-v1"
            },
            {
                "type": "relay-v1"
            },
        ]

    @inlineCallbacks
    def get_connection_hints(self):
        hints = []
        direct_hints = yield self._get_direct_hints()
        for dh in direct_hints:
            hints.append({
                "type": "direct-tcp-v1",
                "priority": dh.priority,
                "hostname": dh.hostname,
                "port": dh.port,  # integer
            })
        for relay in self._transit_relays:
            rhint = {"type": "relay-v1", "hints": []}
            for rh in relay.hints:
                rhint["hints"].append({
                    "type": "direct-tcp-v1",
                    "priority": rh.priority,
                    "hostname": rh.hostname,
                    "port": rh.port
                })
            hints.append(rhint)
        return hints

    def _get_direct_hints(self):
        if self._listener:
            return defer.succeed(self._my_direct_hints)
        # there is a slight race here: if someone calls get_direct_hints() a
        # second time, before the listener has actually started listening,
        # then they'll get a Deferred that fires (with the hints) before the
        # listener starts listening. But most applications won't call this
        # multiple times, and the race is between 1: the parent Wormhole
        # protocol getting the connection hints to the other end, and 2: the
        # listener being ready for connections, and I'm confident that the
        # listener will win.
        self._my_direct_hints, self._listener = self._build_listener()

        if self._listener is None:  # don't listen
            self._listener_d = None
            return defer.succeed(self._my_direct_hints)  # empty

        # Start the server, so it will be running by the time anyone tries to
        # connect to the direct hints we return.
        f = InboundConnectionFactory(self)
        self._listener_f = f  # for tests # XX move to __init__ ?
        self._listener_d = f.whenDone()
        d = self._listener.listen(f)

        def _listening(lp):
            # lp is an IListeningPort
            # self._listener_port = lp # for tests
            def _stop_listening(res):
                lp.stopListening()
                return res

            self._listener_d.addBoth(_stop_listening)
            return self._my_direct_hints

        d.addCallback(_listening)
        return d

    def _stop_listening(self):
        # this is for unit tests. The usual control flow (via connect())
        # wires the listener's Deferred into a there_can_be_only_one(), which
        # eats the errback. If we don't ever call connect(), we must catch it
        # ourselves.
        self._listener_d.addErrback(lambda f: None)
        self._listener_d.cancel()

    def add_connection_hints(self, hints):
        for h in hints:  # hint structs
            hint_type = h.get("type", "")
            if hint_type in ["direct-tcp-v1", "tor-tcp-v1"]:
                dh = parse_tcp_v1_hint(h)
                if dh:
                    self._their_direct_hints.append(dh)  # hint_obj
            elif hint_type == "relay-v1":
                # TODO: each relay-v1 clause describes a different relay,
                # with a set of equally-valid ways to connect to it. Treat
                # them as separate relays, instead of merging them all
                # together like this.
                relay_hints = []
                for rhs in h.get("hints", []):
                    h = parse_tcp_v1_hint(rhs)
                    if h:
                        relay_hints.append(h)
                if relay_hints:
                    rh = RelayV1Hint(hints=tuple(sorted(relay_hints)))
                    self._our_relay_hints.add(rh)
            else:
                log.msg(f"unknown hint type: {h!r}")

    def _send_this(self):
        assert self._transit_key
        if self.is_sender:
            return build_sender_handshake(self._transit_key)
        else:
            return build_receiver_handshake(self._transit_key)

    def _expect_this(self):
        assert self._transit_key
        if self.is_sender:
            return build_receiver_handshake(self._transit_key)
        else:
            return build_sender_handshake(self._transit_key)  # + b"go\n"

    def _sender_record_key(self):
        assert self._transit_key
        if self.is_sender:
            return HKDF(
                self._transit_key,
                SecretBox.KEY_SIZE,
                CTXinfo=b"transit_record_sender_key")
        else:
            return HKDF(
                self._transit_key,
                SecretBox.KEY_SIZE,
                CTXinfo=b"transit_record_receiver_key")

    def _receiver_record_key(self):
        assert self._transit_key
        if self.is_sender:
            return HKDF(
                self._transit_key,
                SecretBox.KEY_SIZE,
                CTXinfo=b"transit_record_receiver_key")
        else:
            return HKDF(
                self._transit_key,
                SecretBox.KEY_SIZE,
                CTXinfo=b"transit_record_sender_key")

    def set_transit_key(self, key):
        assert isinstance(key, bytes), type(key)
        # We use pubsub to protect against the race where the sender knows
        # the hints and the key, and connects to the receiver's transit
        # socket before the receiver gets the relay message (and thus the
        # key).
        self._transit_key = key
        waiters = self._waiting_for_transit_key
        del self._waiting_for_transit_key
        for d in waiters:
            # We don't need eventual-send here. It's safer in general, but
            # set_transit_key() is only called once, and _get_transit_key()
            # won't touch the subscribers list once the key is set.
            d.callback(key)

    def _get_transit_key(self):
        if self._transit_key:
            return defer.succeed(self._transit_key)
        d = defer.Deferred()
        self._waiting_for_transit_key.append(d)
        return d

    @inlineCallbacks
    def connect(self):
        with self._timing.add("transit connect"):
            yield self._get_transit_key()
            # we want to have the transit key before starting any outbound
            # connections, so those connections will know what to say when
            # they connect
            winner = yield self._connect()
        return winner

    def _connect(self):
        # It might be nice to wire this so that a failure in the direct hints
        # causes the relay hints to be used right away (fast failover). But
        # none of our current use cases would take advantage of that: if we
        # have any viable direct hints, then they're either going to succeed
        # quickly or hang for a long time.
        contenders = []
        if self._listener_d:
            contenders.append(self._listener_d)
        relay_delay = 0

        for hint_obj in self._their_direct_hints:
            # Check the hint type to see if we can support it (e.g. skip
            # onion hints on a non-Tor client). Do not increase relay_delay
            # unless we have at least one viable hint.
            ep = endpoint_from_hint_obj(hint_obj, self._tor, self._reactor)
            if not ep:
                continue
            d = self._start_connector(ep,
                                      describe_hint_obj(hint_obj, False, self._tor))
            contenders.append(d)
            relay_delay = self.RELAY_DELAY

        # Start trying the relays a few seconds after we start to try the
        # direct hints. The idea is to prefer direct connections, but not be
        # afraid of using a relay when we have direct hints that don't
        # resolve quickly. Many direct hints will be to unused local-network
        # IP addresses, which won't answer, and would take the full TCP
        # timeout (30s or more) to fail.

        prioritized_relays = {}
        for rh in self._our_relay_hints:
            for hint_obj in rh.hints:
                priority = hint_obj.priority
                if priority not in prioritized_relays:
                    prioritized_relays[priority] = set()
                prioritized_relays[priority].add(hint_obj)

        for priority in sorted(prioritized_relays, reverse=True):
            for hint_obj in prioritized_relays[priority]:
                ep = endpoint_from_hint_obj(hint_obj, self._tor, self._reactor)
                if not ep:
                    continue
                d = task.deferLater(
                    self._reactor,
                    relay_delay,
                    self._start_connector,
                    ep,
                    describe_hint_obj(hint_obj, True, self._tor),
                    is_relay=True)
                contenders.append(d)
            relay_delay += self.RELAY_DELAY

        if not contenders:
            raise TransitError("No contenders for connection")

        winner = there_can_be_only_one(contenders)
        return self._not_forever(2 * TIMEOUT, winner)

    def _not_forever(self, timeout, d):
        """If the timer fires first, cancel the deferred. If the deferred fires
        first, cancel the timer."""
        t = self._reactor.callLater(timeout, d.cancel)

        def _done(res):
            if t.active():
                t.cancel()
            return res

        d.addBoth(_done)
        return d

    def _build_relay_handshake(self):
        return build_sided_relay_handshake(self._transit_key, self._side)

    def _start_connector(self, ep, description, is_relay=False):
        relay_handshake = None
        if is_relay:
            assert self._transit_key
            relay_handshake = self._build_relay_handshake()
        f = OutboundConnectionFactory(self, relay_handshake, description)
        d = ep.connect(f)
        # fires with protocol, or ConnectError
        d.addCallback(lambda p: p.startNegotiation())
        return d

    def connection_ready(self, p):
        # inbound/outbound Connection protocols call this when they finish
        # negotiation. The first one wins and gets a "go". Any subsequent
        # ones lose and get a "nevermind" before being closed.

        if not self.is_sender:
            return "wait-for-decision"

        if self._winner:
            # we already have a winner, so this one loses
            return "nevermind"
        # this one wins!
        self._winner = p
        return "go"


class TransitSender(Common):
    is_sender = True


class TransitReceiver(Common):
    is_sender = False


# based on twisted.protocols.ftp.FileConsumer, but don't close the filehandle
# when done, and add a progress function that gets called with the length of
# each write, and a hasher function that gets called with the data.


@implementer(interfaces.IConsumer)
class FileConsumer:
    def __init__(self, f, progress=None, hasher=None):
        self._f = f
        self._progress = progress
        self._hasher = hasher
        self._producer = None

    def registerProducer(self, producer, streaming):
        assert not self._producer
        self._producer = producer
        assert streaming

    def write(self, bytes):
        self._f.write(bytes)
        if self._progress:
            self._progress(len(bytes))
        if self._hasher:
            self._hasher(bytes)

    def unregisterProducer(self):
        assert self._producer
        self._producer = None


# the TransitSender/Receiver.connect() yields a Connection, on which you can
# do send_record(), but what should the receive API be? set a callback for
# inbound records? get a Deferred for the next record? The producer/consumer
# API is enough for file transfer, but what would other applications want?

# how should the Listener be managed? we want to shut it down when the
# connect() Deferred is cancelled, as well as terminating any negotiations in
# progress.
#
# the factory should return/manage a deferred, which fires iff an inbound
# connection completes negotiation successfully, can be cancelled (which
# stops the listener and drops all pending connections), but will never
# timeout, and only errbacks if cancelled.

# write unit test for _ThereCanBeOnlyOne

# check start/finish time-gathering instrumentation

# relay URLs are probably mishandled: both sides probably send their URL,
# then connect to the *other* side's URL, when they really should connect to
# both their own and the other side's. The current implementation probably
# only works if the two URLs are the same.



from unittest import mock
from zope.interface import alsoProvides
from ..._interfaces import IDilationManager
from ..._dilation.connection import Open, Data, Close
from ..._dilation.inbound import (Inbound, DuplicateOpenError,
                                  DataForMissingSubchannelError,
                                  CloseForMissingSubchannelError)


def make_inbound():
    m = mock.Mock()
    alsoProvides(m, IDilationManager)
    m._subprotocol_factories = mock.Mock()
    host_addr = object()
    i = Inbound(m, host_addr)
    return i, m, host_addr


def test_seqnum():
    i, m, host_addr = make_inbound()
    r1 = Open(scid=513, seqnum=1, subprotocol="proto")
    r2 = Data(scid=513, seqnum=2, data=b"")
    r3 = Close(scid=513, seqnum=3)
    assert not i.is_record_old(r1)
    assert not i.is_record_old(r2)
    assert not i.is_record_old(r3)

    i.update_ack_watermark(r1.seqnum)
    assert i.is_record_old(r1)
    assert not i.is_record_old(r2)
    assert not i.is_record_old(r3)

    i.update_ack_watermark(r2.seqnum)
    assert i.is_record_old(r1)
    assert i.is_record_old(r2)
    assert not i.is_record_old(r3)


def test_open_data_close(observe_errors):
    i, m, host_addr = make_inbound()
    scid1 = b"scid"
    scid2 = b"scXX"
    c = mock.Mock()
    i.use_connection(c)
    sc1 = mock.Mock()
    peer_addr = object()
    with mock.patch("wormhole._dilation.inbound.SubChannel",
                    side_effect=[sc1]) as sc:
        with mock.patch("wormhole._dilation.inbound.SubchannelAddress",
                        side_effect=[peer_addr]) as sca:
            i.handle_open(scid1, "proto")
    assert m._subprotocol_factories.mock_calls == [mock.call._got_open(sc1, peer_addr)]
    assert sc.mock_calls == [mock.call(scid1, m, host_addr, peer_addr)]
    assert sca.mock_calls == [mock.call("proto")]

    # reset calls
    m._subprotocol_factories.mock_calls[:] = []

    # a subsequent duplicate OPEN should be ignored
    with mock.patch("wormhole._dilation.inbound.SubChannel",
                    side_effect=[sc1]) as sc:
        with mock.patch("wormhole._dilation.inbound.SubchannelAddress",
                        side_effect=[peer_addr]) as sca:
            i.handle_open(scid1, "proto")
    assert m._subprotocol_factories.mock_calls == []
    assert sc.mock_calls == []
    assert sca.mock_calls == []
    observe_errors.flush(DuplicateOpenError)

    i.handle_data(scid1, b"data")
    assert sc1.mock_calls == [mock.call.remote_data(b"data")]
    sc1.mock_calls[:] = []

    i.handle_data(scid2, b"for non-existent subchannel")
    assert sc1.mock_calls == []
    observe_errors.flush(DataForMissingSubchannelError)

    i.handle_close(scid1)
    assert sc1.mock_calls == [mock.call.remote_close()]
    sc1.mock_calls[:] = []

    i.handle_close(scid2)
    assert sc1.mock_calls == []
    observe_errors.flush(CloseForMissingSubchannelError)

    # after the subchannel is closed, the Manager will notify Inbound
    i.subchannel_closed(scid1, sc1)

    i.stop_using_connection()


def test_pause():
    i, m, host_addr = make_inbound()
    c = mock.Mock()

    # add two subchannels, pause one, then add a connection
    scid1 = b"sci1"
    scid2 = b"sci2"
    sc1 = mock.Mock()
    sc2 = mock.Mock()
    peer_addr = object()
    with mock.patch("wormhole._dilation.inbound.SubChannel",
                    side_effect=[sc1, sc2]):
        with mock.patch("wormhole._dilation.inbound.SubchannelAddress",
                        return_value=peer_addr):
            i.handle_open(scid1, "proto")
            i.handle_open(scid2, "proto")
    assert c.mock_calls == []

    i.subchannel_pauseProducing(sc1)
    assert c.mock_calls == []
    i.subchannel_resumeProducing(sc1)
    assert c.mock_calls == []
    i.subchannel_pauseProducing(sc1)
    assert c.mock_calls == []

    i.use_connection(c)
    assert c.mock_calls == [mock.call.pauseProducing()]
    c.mock_calls[:] = []

    i.subchannel_resumeProducing(sc1)
    assert c.mock_calls == [mock.call.resumeProducing()]
    c.mock_calls[:] = []

    # consumers aren't really supposed to do this, but tolerate it
    i.subchannel_resumeProducing(sc1)
    assert c.mock_calls == []

    i.subchannel_pauseProducing(sc1)
    assert c.mock_calls == [mock.call.pauseProducing()]
    c.mock_calls[:] = []
    i.subchannel_pauseProducing(sc2)
    assert c.mock_calls == []  # was already paused

    # tolerate duplicate pauseProducing
    i.subchannel_pauseProducing(sc2)
    assert c.mock_calls == []

    # stopProducing is treated like a terminal resumeProducing
    i.subchannel_stopProducing(sc1)
    assert c.mock_calls == []
    i.subchannel_stopProducing(sc2)
    assert c.mock_calls == [mock.call.resumeProducing()]
    c.mock_calls[:] = []



from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces


@implementer(_interfaces.ITerminator)
class Terminator:
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __init__(self):
        self._mood = None

    def wire(self, boss, rendezvous_connector, nameplate, mailbox, dilator):
        self._B = _interfaces.IBoss(boss)
        self._RC = _interfaces.IRendezvousConnector(rendezvous_connector)
        self._N = _interfaces.INameplate(nameplate)
        self._M = _interfaces.IMailbox(mailbox)
        self._D = _interfaces.IDilator(dilator)

    # 2*2-1+1 main states:
    # (nm, m, n, d): nameplate and/or mailbox is active
    # (o, ""): open (not-yet-closing), or trying to close
    # after closing the mailbox-server connection, we stop Dilation
    # S0 is special: we don't hang out in it

    # TODO: rename o to 0, "" to 1. "S1" is special/terminal
    # so S0nm/S0n/S0m/S0, S1nm/S1n/S1m/(S1)

    # We start in Snmo (non-closing). When both nameplate and mailboxes are
    # done, and we're closing, then we stop the RendezvousConnector

    @m.state(initial=True)
    def Snmo(self):
        pass  # pragma: no cover

    @m.state()
    def Smo(self):
        pass  # pragma: no cover

    @m.state()
    def Sno(self):
        pass  # pragma: no cover

    @m.state()
    def S0o(self):
        pass  # pragma: no cover

    @m.state()
    def Snm(self):
        pass  # pragma: no cover

    @m.state()
    def Sm(self):
        pass  # pragma: no cover

    @m.state()
    def Sn(self):
        pass  # pragma: no cover

    # @m.state()
    # def S0(self): pass # unused

    @m.state()
    def S_stoppingRC(self):
        pass  # pragma: no cover

    @m.state()
    def S_stoppingD(self):
        pass  # pragma: no cover

    @m.state()
    def S_stopped(self, terminal=True):
        pass  # pragma: no cover

    # from Boss
    @m.input()
    def close(self, mood):
        pass

    # from Nameplate
    @m.input()
    def nameplate_done(self):
        pass

    # from Mailbox
    @m.input()
    def mailbox_done(self):
        pass

    # from RendezvousConnector
    @m.input()
    def stoppedRC(self):
        pass

    @m.input()
    def stoppedD(self):
        pass

    @m.output()
    def close_nameplate(self, mood):
        self._N.close()  # ignores mood

    @m.output()
    def close_mailbox(self, mood):
        self._M.close(mood)

    @m.output()
    def ignore_mood_and_RC_stop(self, mood):
        self._RC.stop()

    @m.output()
    def RC_stop(self):
        self._RC.stop()

    @m.output()
    def stop_dilator(self):
        self._D.stop()

    @m.output()
    def B_closed(self):
        self._B.closed()

    Snmo.upon(mailbox_done, enter=Sno, outputs=[])
    Snmo.upon(close, enter=Snm, outputs=[close_nameplate, close_mailbox])
    Snmo.upon(nameplate_done, enter=Smo, outputs=[])

    Sno.upon(close, enter=Sn, outputs=[close_nameplate])
    Sno.upon(nameplate_done, enter=S0o, outputs=[])

    Smo.upon(close, enter=Sm, outputs=[close_mailbox])
    Smo.upon(mailbox_done, enter=S0o, outputs=[])

    Snm.upon(mailbox_done, enter=Sn, outputs=[])
    Snm.upon(nameplate_done, enter=Sm, outputs=[])

    Sn.upon(nameplate_done, enter=S_stoppingRC, outputs=[RC_stop])
    Sm.upon(mailbox_done, enter=S_stoppingRC, outputs=[RC_stop])
    S0o.upon(close, enter=S_stoppingRC, outputs=[ignore_mood_and_RC_stop])

    S_stoppingRC.upon(stoppedRC, enter=S_stoppingD, outputs=[stop_dilator])

    S_stoppingD.upon(stoppedD, enter=S_stopped, outputs=[B_closed])
